@@ -56,8 +56,6 @@ struct phdr {
 
 #define MAXSOCKS	32
 
-extern int port;
-
 static int synsock[MAXSOCKS];
 static int current;
 
@@ -206,7 +204,7 @@ synprobe_send(struct in_addr dst, u_int32_t *seqnr)
 
 	tcp = (struct tcphdr *)(buf + sizeof(*ip));
 	tcp->th_sport = ntohs(*seqnr);
-	tcp->th_dport = ntohs(port);
+	tcp->th_dport = ntohs(22);
 	tcp->th_off = 5;
 	tcp->th_flags = TH_SYN;
 	tcp->th_win = htons(16384);
