@@ -1,5 +1,5 @@
 /*
- * ScanSSH - simple SSH version scanner V1.5
+ * ScanSSH - simple SSH version scanner
  *
  * Copyright 2000-2001 Niels Provos <provos@citi.umich.edu>
  * All rights reserved.
@@ -48,6 +48,8 @@
 #include <unistd.h>
 #include <md5.h>
 
+#define VERSION	"V1.55"
+
 #include "config.h"
 #ifdef HAVE_FDMASK_IN_SELECT
 #include <sys/select.h>
@@ -97,21 +99,22 @@ struct generate {
 	u_int32_t gen_max;
 };
 
-int populate(struct argument *, int);
+int populate(struct argument **, int *);
 int next_address(struct generate *, struct in_addr *);
 
 /* Globals */
 int usepcap;
+int ssh_sendident;
 pcap_t *pd;
 int rndexclude = 1;
 
+struct address_slot slots[MAXSLOTS];
+
 #define MAX_PROCESSES	30
 
 int commands[MAX_PROCESSES];
 int results[MAX_PROCESSES];
 
-struct argument *args;
-
 TAILQ_HEAD (gen_list, generate) genqueue;
 TAILQ_HEAD (queue_list, argument) readyqueue;
 TAILQ_HEAD (syn_list, argument) synqueue;
@@ -121,6 +124,43 @@ TAILQ_HEAD (syn_list, argument) synqueue;
 
 int synqueuesz;
 
+struct address_slot *
+slot_get(void)
+{
+	int i;
+	struct address_slot *slot;
+
+	for (i = 0; i < MAXSLOTS; i++)
+		if (slots[i].slot_base == NULL)
+			break;
+
+	if (i >= MAXSLOTS)
+		return (NULL);
+	
+	slot = &slots[i];
+
+	if (slot->slot_base == NULL) {
+		slot->slot_size = EXPANDEDARGS;
+		slot->slot_base = xmalloc(EXPANDEDARGS * sizeof(struct argument));
+		memset(slot->slot_base, 0,
+		       slot->slot_size * sizeof(struct argument));
+	}
+
+	return (slot);
+}
+
+void
+slot_free(struct address_slot *slot)
+{
+	slot->slot_ref--;
+	if (slot->slot_ref)
+		return;
+
+	slot->slot_size = 0;
+	free(slot->slot_base);
+	slot->slot_base = 0;
+}
+
 void
 synlist_init(void)
 {
@@ -139,6 +179,7 @@ synlist_insert(struct argument *arg)
 
 	timerclear(&tmp);
 	tmp.tv_sec += (arg->a_retry/2 + 1) * SYNWAIT;
+	tmp.tv_usec = arc4random();
 
 	gettimeofday(&arg->a_tv, NULL);
 	timeradd(&tmp, &arg->a_tv, &arg->a_tv);
@@ -226,13 +267,14 @@ printres(struct argument *exp, char *result)
 }
 
 void
-scanips(struct argument *args, u_int32_t nargs)
+scanips(void)
 {
 	u_int32_t entries = 0;
 	fd_set *readset, *writeset;
 	int maxfd = 0, fdsetsz, counts = 0;
 	struct timeval tv;
 	int res, i, lastinsert, newinsert;
+	struct argument *args;
 	char buf[MAXBUF], *p;
 
 	for (i = 0; i < MAX_PROCESSES; i++) {
@@ -256,8 +298,10 @@ scanips(struct argument *args, u_int32_t nargs)
 	while (TAILQ_FIRST(&genqueue) || entries ||
 	       counts || !synlist_empty()) {
 		/* Create new entries, if we need them */
-		if (!entries)
-			entries = populate(args, nargs);
+		if (!entries && TAILQ_FIRST(&genqueue)) {
+			 if (populate(&args, &entries) == -1)
+				 entries = 0;
+		}
 
 		memset(readset, 0, fdsetsz);
 		memset(writeset, 0, fdsetsz);
@@ -374,8 +418,10 @@ scanips(struct argument *args, u_int32_t nargs)
 						break;
 					}
 					nsent++;
-				} else
+				} else {
 					printres(exp, "<timeout>");
+					slot_free(exp->a_slot);
+				}
 			}
 
 			/* Put new entries in the synlist */
@@ -420,6 +466,7 @@ scanips(struct argument *args, u_int32_t nargs)
 				TAILQ_REMOVE(&readyqueue, arg, a_next);
 				atomicio(write, commands[i], arg,
 					 sizeof(struct argument));
+				slot_free(arg->a_slot);
 				counts++;
 				newinsert = i;
 			} else if (synlist_empty()) {
@@ -483,7 +530,9 @@ void
 usage(char *name)
 {
 	fprintf(stderr, 
-		"%s: [-ERh] [-e excludefile] [-p ifaddr] <IP address|network>...\n\n"
+		"%s: [-VIERh] [-e excludefile] [-p ifaddr] <IP address|network>...\n\n"
+		"\t-V          print version number of scanssh,\n"
+		"\t-I          do not send identification string,\n"
 		"\t-E          exit if exclude file is missing,\n"
 		"\t-R          do not honor exclude file for random addresses,\n"
 		"\t-e <file>   exclude the IP addresses and networks in <file>,\n"
@@ -549,12 +598,24 @@ generate_free(struct generate *gen)
  */
 
 int
-populate(struct argument *args, int count)
+populate(struct argument **pargs, int *nargs)
 {
 	struct generate *gen;
 	struct in_addr addr;
+	struct address_slot *slot = NULL;
+	struct argument *args;
+	int count;
+
 	u_int32_t i = 0;
 
+	if (!TAILQ_FIRST(&genqueue))
+		return (-1);
+
+	slot = slot_get();
+
+	args = slot->slot_base;
+	count = slot->slot_size;
+
 	while (TAILQ_FIRST(&genqueue) && count) {
 		gen = TAILQ_FIRST(&genqueue);
 		
@@ -574,13 +635,19 @@ populate(struct argument *args, int count)
 
 			args[i].a_type = AF_INET;
 			args[i].a_ipv4.s_addr = htonl(addr.s_addr);
+			args[i].a_slot = slot;
+
+			slot->slot_ref++;
 
 			count--;
 			i++;
 		}
 	}
 
-	return (i);
+	*pargs = args;
+	*nargs = i;
+
+	return (0);
 }
 
 int
@@ -943,12 +1010,19 @@ main(int argc, char **argv)
 	char *name, *src = NULL;
 	int ch;
 	struct argument *args = NULL;
-	u_int32_t nargs;
 	int failonexclude = 0;
 
+	ssh_sendident = 1;
+
 	name = argv[0];
-	while ((ch = getopt(argc, argv, "hp:e:ER")) != -1)
+	while ((ch = getopt(argc, argv, "VIhp:e:ER")) != -1)
 		switch(ch) {
+		case 'V':
+			fprintf(stderr, "ScanSSH %s\n", VERSION);
+			exit(0);
+		case 'I':
+			ssh_sendident = 0;
+			break;
 		case 'p':
 			src = optarg;
 			break;
@@ -993,9 +1067,7 @@ main(int argc, char **argv)
 		exit(1);
 	}
 
-	nargs = EXPANDEDARGS;
-	args = xmalloc(nargs * sizeof(struct argument));
-	memset(args, 0, nargs * sizeof(struct argument)); 
+	memset(slots, 0, sizeof(slots));
 
 	TAILQ_INIT(&genqueue);
 
@@ -1010,7 +1082,7 @@ main(int argc, char **argv)
 	if (!TAILQ_FIRST(&genqueue))
 		errx(1, "nothing to scan");
 
-	scanips(args, nargs);
+	scanips();
 
 	return (1);
 }
